/*$6 ***************************************************************************
Copyright (C) <2007> <Sven Trenkel, Rene Weller, Gabriel Zachmann>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 ******************************************************************************* */
/**
 *  Implementation of the 3 methods for configuration computing: Grid method,
 *  the sphere method and the advanced sphere method called csphere
 *  See http://cg.in.tu-clausthal.de/papers/wscg07/wscg_benchmarking_suite.pdf
 *  for details
 *
 *  \author Sven Trenkel, Rene Weller, Gabriel Zachmann
 */
 
#include <OpenSG/OSGConfig.h>

#include <OpenSG/OSGBaseTypes.h>
#include <OpenSG/OSGNodePtr.h>
#include <OpenSG/OSGTransform.h>

#include "Bench.h"
#include "Grid.h"

#define min( a, b ) ( (a) < (b) ? (a) : (b) )
#define max( a, b ) ( (a) > (b) ? (a) : (b) )

using osg:: Pnt3f;
using osg:: Vec3f;

/*! ----------------------------------------------------------------------------
 *  \brief  Do one iteration. Calling this method will queue the next position
 *          for testing. The positions are generated by positioning the center
 *          of one object at a great distance from the other. Then it is moved
 *          closer to the center of the other object until the distance of both
 *          objects is right. The initial positions are generated by a sphere
 *          around the fixed object with longitude/latitude coordinates. The
 *          first position is at -90/0.
 *  \return true if the next position
 *          will reset the scene to its initial position, otherwise false.
 ------------------------------------------------------------------------------- */
bool Sphere::iterate()
{
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
    static const osg::Pnt3f nullPnt( 0, 0, 0 );
    double                  diff = settings->dist;
    double                  xz_rad = xzDeg / 180.0 * 3.1415926538;
    double                  yz_rad = yzDeg / 180.0 * 3.1415926538;
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

    /* Generate the current position */
    if( diff < 0 )
        diff = 0;
    osg::Vec3f pos( cos( xz_rad ) * cos( yz_rad ) * (fixedSize.length() +
                    movingSize.length() + diff + 0.1) + fixedCenter[0] - movingCenter[0], sin( yz_rad ) * (fixedSize.length() +
                    movingSize.length() + diff + 0.1) + fixedCenter[1] -
                    movingCenter[1], sin( xz_rad ) * cos( yz_rad ) * (
                        fixedSize.length() + movingSize.length() + diff + 0.1) +
                    fixedCenter[2] - movingCenter[2] );
    queueBetween( nullPnt, pos );

    /*
     * Increase the xz position. If it's greater than 360 reset it to 0 and
     * increase the yz position.
     */
    xzDeg += settings->degVel;
    if( xzDeg >= 360 || yzDeg == -90 )
    {
        xzDeg = 0;
        yzDeg += settings->degVel;
        if( yzDeg > 90 )
        {
            yzDeg = -90;
            return true;
        }
    }
    else if( yzDeg == 90 )
    {
        xzDeg = 0;
        yzDeg = -90;
        return true;
    }

    return false;
}

/*! ----------------------------------------------------------------------------
 *  \brief  Do one iteration. Calling this method will queue the next position
 *          for testing. The gerneration of positions is similar to
 *          Sphere::iterate() but there is an additional check to make sure
 *          that the generated positions have about the same distance to each
 *          other.
 * \return true if the next position will reset the scene to
 *          its initial position, otherwise false.
 ------------------------------------------------------------------------------- */
bool CheckSphere::iterate()
{
    static const osg::Pnt3f nullPnt( 0, 0, 0 );
    static osg::Pnt3f       lastPos( 0, 0, 0 ), lastYPos;
    osg::Pnt3f              ret;
    double                  diff = settings->dist;
    double                  xz_rad;
    double                  yz_rad;

    double                  best_dist = 0, best_deg = xzDeg;

    double                  min_deg = xzDeg -
        settings->degVel /
        2, max_deg = xzDeg;

    if( diff < 0 )
        diff = 0;
    xz_rad = xzDeg / 180.0 * 3.1415926538;
    yz_rad = yzDeg / 180.0 * 3.1415926538;
    osg::Pnt3f pos( cos( xz_rad ) * cos( yz_rad ) * (fixedSize.length() +
                    movingSize.length() + diff + 0.1) + fixedCenter[0] - movingCenter[0], sin( yz_rad ) * (fixedSize.length() +
                    movingSize.length() + diff + 0.1) + fixedCenter[1] -
                    movingCenter[1], sin( xz_rad ) * cos( yz_rad ) * (
                        fixedSize.length() + movingSize.length() + diff + 0.1) +
                    fixedCenter[2] - movingCenter[2] );
    if( testBetween( nullPnt, pos, ret ) )
        queue( ret );
    lastPos = ret;
    if( !xzDeg )
        lastYPos = ret;

    if( yzDeg == 90 )
    {
        xzDeg = 0;
        yzDeg = -90;
        return true;
    }

    do
    {
        xzDeg += settings->degVel / 2;
        if( xzDeg >= 360 || yzDeg == -90 )
            break;

        xz_rad = xzDeg / 180.0 * 3.1415926538;
        yz_rad = yzDeg / 180.0 * 3.1415926538;
        osg::Vec3f pos( cos( xz_rad ) * cos( yz_rad ) * (fixedSize.length() +
                        movingSize.length() + diff + 0.1) + fixedCenter[0] - movingCenter[0], sin( yz_rad ) * (fixedSize.length() +
                        movingSize.length() + diff + 0.1) + fixedCenter[1] -
                        movingCenter[1], sin( xz_rad ) * cos( yz_rad ) * (
                            fixedSize.length() + movingSize.length() + diff +
                        0.1) + fixedCenter[2] - movingCenter[2] );
        testBetween( nullPnt, pos, ret );
        if( fabs( (ret - lastPos).length() - avDist ) < best_dist )
        {
            best_dist = fabs( (ret - lastPos).length() - avDist );
            best_deg = xzDeg;
        }
    } while( (ret - lastPos).length() < avDist * 0.8 );

    while( (ret - lastPos).length() > avDist * 1.2 )
    {
        xzDeg = ( min_deg + max_deg ) / 2;
        if( xzDeg >= 360 || yzDeg == -90 )
            break;

        xz_rad = xzDeg / 180.0 * 3.1415926538;
        yz_rad = yzDeg / 180.0 * 3.1415926538;
        osg::Vec3f pos( cos( xz_rad ) * cos( yz_rad ) * (fixedSize.length() +
                        movingSize.length() + diff + 0.1) + fixedCenter[0] - movingCenter[0], sin( yz_rad ) * (fixedSize.length() +
                        movingSize.length() + diff + 0.1) + fixedCenter[1] -
                        movingCenter[1], sin( xz_rad ) * cos( yz_rad ) * (
                            fixedSize.length() + movingSize.length() + diff +
                        0.1) + fixedCenter[2] - movingCenter[2] );
        testBetween( nullPnt, pos, ret );
        if( fabs( (ret - lastPos).length() - avDist ) < best_dist )
        {
            best_dist = fabs( (ret - lastPos).length() - avDist );
            best_deg = xzDeg;
        }

        if( (ret - lastPos).length() > avDist )
            max_deg = xzDeg;
        else
            min_deg = xzDeg;
        if( max_deg - min_deg < 0.5 )
            break;
    }

    if( xzDeg >= 360 || yzDeg == -90 )
    {
        xzDeg = 0;
        yzDeg += settings->degVel;
        xz_rad = xzDeg / 180.0 * 3.1415926538;
        yz_rad = yzDeg / 180.0 * 3.1415926538;
        osg::Vec3f pos( cos( xz_rad ) * cos( yz_rad ) * (fixedSize.length() +
                        movingSize.length() + diff + 0.1) + fixedCenter[0] - movingCenter[0], sin( yz_rad ) * (fixedSize.length() +
                        movingSize.length() + diff + 0.1) + fixedCenter[1] -
                        movingCenter[1], sin( xz_rad ) * cos( yz_rad ) * (
                            fixedSize.length() + movingSize.length() + diff +
                        0.1) + fixedCenter[2] - movingCenter[2] );
        testBetween( nullPnt, pos, ret );
        if( lastPos != nullPnt && avDist == -1 )
            avDist = ( ret - lastPos ).length();
        if( yzDeg > 90 )
        {
            yzDeg = -90;
            return true;
        }
    }

    return false;
}

/*! ----------------------------------------------------------------------------
 *  \brief  test a position of the grid. Determines the distance of the field
 *          of the grid at (x, y, z) and updates the information of the fields.
 ------------------------------------------------------------------------------- */
void Grid::test( int x, int y, int z )
{
    double  dist = testPos( field[x][y][z].pos );
    updatePosition( x, y, z, dist );
}

/*! ----------------------------------------------------------------------------
 *  \brief  Do one iteration. Calling this method will queue the next position
 *          for testing. The center of the moving object is positioned in the
 *          center of a cell of the grid and the position of the objects is
 *          calculated. The the distance is greater than the desired distance
 *          and the distance in a neighbouring cell was smaller, or the other
 *          way round, the correct position between these points is queued.
 *  \return true if the next position will reset the scene to its
 *          initial position, otherwise false.
 ------------------------------------------------------------------------------- */
bool Grid::iterate()
{
    double  dist = settings->dist >= 0 ? settings->dist : 0;

    for( ; x <= size[0]; ++x )
    {
        for( ; y <= size[1]; ++y )
        {
            for( ; z <= size[2]; ++z )
            {
                if( field[x][y][z].donttest )
                    continue;
                if( State <= 0 )
                {
                    State = 1;
                    test( x, y, z );
                }

                if( State <= 1 && z > 0 && field[x][y][z - 1].tested && ((
                        field[x][y][z].distance > dist
                && field[x][y][z - 1].distance <= dist) || (
                        field[x][y][z].distance <= dist
                && field[x][y][z - 1].distance > dist)) )
                {
                    State = 2;
                    queueBetween( field[x][y][z].pos, field[x][y][z - 1].pos );
                    return false;
                }

                if( State <= 2 && y > 0 && field[x][y - 1][z].tested && ((
                        field[x][y][z].distance > dist
                && field[x][y - 1][z].distance <= dist) || (
                        field[x][y][z].distance <= dist
                && field[x][y - 1][z].distance > dist)) )
                {
                    State = 3;
                    queueBetween( field[x][y][z].pos, field[x][y - 1][z].pos );
                    return false;
                }

                if( State <= 3 && x > 0 && field[x - 1][y][z].tested && ((
                        field[x][y][z].distance > dist
                && field[x - 1][y][z].distance <= dist) || (
                        field[x][y][z].distance <= dist
                && field[x - 1][y][z].distance > dist)) )
                {
                    State = 4;
                    queueBetween( field[x][y][z].pos, field[x - 1][y][z].pos );
                    return false;
                }

                State = 0;
            }

            z = 0;
        }

        y = 0;
    }

    x = 0;
    return true;
}

/* -----------------------------------------------------------------------------
 *  \brief Set distance of grid cell x,y,z to distance dist
 ------------------------------------------------------------------------------- */
void Grid::updatePosition( int x, int y, int z, double dist )
{
    int         set = 0;
    static int  sum = 0;
    double      radius = fabs( dist - settings->dist );
    int         fielddiff[3] = { int( radius / fieldSize[0] ) - 2, int( radius / fieldSize[1] ) - 2, int( radius / fieldSize[2] ) - 2 };
    field[x][y][z].setStatus( dist );
    
    int         minX = max( 0, x - fielddiff[0] ), maxX = min( x + fielddiff[0],
                                                               size[0] );
    int         minY = max( 0, y - fielddiff[1] ), maxY = min( y + fielddiff[1],
                                                               size[1] );
    int         minZ = max( 0, z - fielddiff[2] ), maxZ = min( z + fielddiff[2],
                                                               size[2] );
                                                               
    if( fielddiff[0] <= 0 || fielddiff[1] <= 0 || fielddiff[2] <= 0 )
        return;

    return;
    for( int i = minX; i <= maxX; ++i )
    {
        for( int j = minY; j <= maxY; ++j )
        {
            for( int k = minZ; k <= maxZ; ++k )
            {
                double  dist = sqrt( (i - x) * (i - x) * fieldSize[0] * fieldSize[0] +
                                         (j - y) * (j - y) * fieldSize[1] * fieldSize[1] +
                                                 (k - z) * (k - z) * fieldSize[2] * fieldSize[2] );
                if( field[i][j][k].tested || field[i][j][k].donttest )
                    continue;
                if( dist > radius )
                    continue;
                field[i][j][z].donttest = true;
                ++set;
                ++sum;
            }
        }
    }
}

/*! ----------------------------------------------------------------------------
 *  \brief  Constructor. Does nothing.
 ------------------------------------------------------------------------------- */
Sphere::Sphere( Settings *set ) : PosFinder( set ), xzDeg( 0 ), yzDeg( -90 )
{
}

/*! ----------------------------------------------------------------------------
 *  \brief  Constructor. Does nothing.
 ------------------------------------------------------------------------------- */
CheckSphere::CheckSphere( Settings *set ) : PosFinder( set ), xzDeg( 0 ), yzDeg( -90 ), avDist( -1 )
{
}

/*! ----------------------------------------------------------------------------
 *  \brief  Constructor. Creates the grid..
 ------------------------------------------------------------------------------- */
Grid::Grid( Settings *set ) : PosFinder( set ), x( 0 ), y( 0 ), z( 0 ), State( 0 ), field( 0 )
{
    const char  *p;
    int         maxsize[3];
    if( !settings->qridSize )
        settings->qridSize = "20";
    maxsize[0] = maxsize[1] = maxsize[2] = atoi( settings->qridSize );
    p = strchr( settings->qridSize, ',' );
    if( p )
    {
        ++p;
        maxsize[1] = maxsize[2] = atoi( p );
        p = strchr( p, ',' );
        if( p )
        {
            ++p;
            maxsize[2] = atoi( p );
        }
    }

    makeGrid( maxsize );
}

/*! ----------------------------------------------------------------------------
 *  \brief  Rotates the fixed object. Public interface for PosFinder::doRot().
 *          It deletes the old grid and creates a new one.
 ------------------------------------------------------------------------------- */
bool Grid::rotate()
{
    bool        ret = doRot();

    const char  *p;
    int         maxsize[3];
    if( !settings->qridSize )
        settings->qridSize = "20";
    maxsize[0] = maxsize[1] = maxsize[2] = atoi( settings->qridSize );
    p = strchr( settings->qridSize, ',' );
    if( p )
    {
        ++p;
        maxsize[1] = maxsize[2] = atoi( p );
        p = strchr( p, ',' );
        if( p )
        {
            ++p;
            maxsize[2] = atoi( p );
        }
    }

    deleteGrid();
    makeGrid( maxsize );

    return ret;
}

/*! ----------------------------------------------------------------------------
 *  \brief  Rotates the fixed object. Public interface for PosFinder::doRot().
 *          Does nothing except calling it.
 ------------------------------------------------------------------------------- */
bool Sphere::rotate()
{
    return doRot();
}

/*! ----------------------------------------------------------------------------
 *  \brief  Rotates the fixed object. Public interface for PosFinder::doRot().
 *          Does nothing except calling it.
 ------------------------------------------------------------------------------- */
bool CheckSphere::rotate()
{
    return doRot();
}

Grid::~Grid()
{
    deleteGrid();
}

/*! ----------------------------------------------------------------------------
 *  \brief  Deletes the grid and frees its memory.
 ------------------------------------------------------------------------------- */
void Grid::deleteGrid()
{
    if( !field )
        return;
    for( int x = 0; x <= size[0]; ++x )
    {
        for( int y = 0; y <= size[1]; ++y )
            delete[] field[x][y];
        delete[] field[x];
    }

    delete[] field;
}

/*! ----------------------------------------------------------------------------
 *  \brief  Create the grid. Allocates the fields of the grid and initilizes
 *          them. 
 *  \param maxsize The size of the grid.
 ------------------------------------------------------------------------------- */
void Grid::makeGrid( int maxsize[3] )
{
    double  dist = settings->dist >= 0 ? settings->dist : 0;
    fieldSize[0] = double( fixedSize[0] + movingSize[0] + dist ) * 2.0 / double( maxsize[0] );
    fieldSize[1] = double( fixedSize[1] + movingSize[1] + dist ) * 2.0 / double( maxsize[1] );
    fieldSize[2] = double( fixedSize[2] + movingSize[2] + dist ) * 2.0 / double( maxsize[2] );
    for( int i = 0; i < 3; ++i )
        size[i] = maxsize[i];

    field = new Field **[size[0] + 1];
    for( int x = 0; x <= size[0]; ++x )
    {
        double  xPos = fixedCenter[0] -
            fixedSize[0] -
            movingCenter[0] -
            movingSize[0] +
            x *
            fieldSize[0] -
            dist;
        field[x] = new Field *[size[1] + 1];
        for( int y = 0; y <= size[1]; ++y )
        {
            double  yPos = fixedCenter[1] -
                fixedSize[1] -
                movingCenter[1] -
                movingSize[1] +
                y *
                fieldSize[1] -
                dist;
            field[x][y] = new Field[size[2] + 1];
            for( int z = 0; z <= size[2]; ++z )
            {
                double  zPos = fixedCenter[2] -
                    fixedSize[2] -
                    movingCenter[2] -
                    movingSize[2] +
                    z *
                    fieldSize[2] -
                    dist;
                field[x][y][z].pos = osg::Pnt3f( xPos, yPos, zPos );
            }
        }
    }
}
