/*$6 ***************************************************************************
Copyright (C) <2007> <Sven Trenkel, Rene Weller, Gabriel Zachmann>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 ******************************************************************************* */

/**
 *  \ Setting \
 */

/**
 *  \class Settings 
 *  \brief Administrates the benchmarking parameters such as
 *  used algorithm, objects... 
 *  \author Sven Trenkel, Rene Weller, Gabriel Zachmann
 */

/*
 * Includes ;
 */
#include <list>

#include <OpenSG/OSGConfig.h>

#include <OpenSG/OSGBaseTypes.h>
#include <OpenSG/OSGBaseFunctions.h>
#include <OpenSG/OSGNodePtr.h>
#include <OpenSG/OSGGeometry.h>
#include <OpenSG/OSGGeoProperty.h>
#include <OpenSG/OSGGeoFunctions.h>
#include <OpenSG/OSGSimpleGeometry.h>
#include <OpenSG/OSGSimpleMaterial.h>
#include <OpenSG/OSGSceneFileHandler.h>
#include <OpenSG/OSGAttachment.h>

#include <OpenSG/OSGSimpleAttachments.h>
#include <OpenSG/OSGSimpleSceneManager.h>
#include <OpenSG/OSGAttachment.h>
#include <OpenSG/OSGGLUTWindow.h>
#include <OpenSG/OSGTrackball.h>
#include <OpenSG/OSGFaceIterator.h>
#include <OpenSG/OSGSimpleMaterial.h>
#include <OpenSG/OSGSceneFileHandler.h>
#include <OpenSG/OSGTriangleIterator.h>

#include "Settings.h"
#include "Utils.h"

using osg::Pnt3f;

/**
 *  Settings ;
 */

/*! ----------------------------------------------------------------------------
 *  \brief  Used to register an algorithm. Set benchmarking parameters to
 *          default
 ------------------------------------------------------------------------------- */
Settings::Settings() : Obj( 0 ), algo( 0 ), curDist( 0 ), curFile( 0 ), saveName( 0 ), qridSize( 0 ), method( METH_SPHERE ), dist( -1 ), rotVel( 60 ), degVel( 30 ), validate( false ), noTiming( false ), withWindow( false ), choice( 0 ), maxminTiming( true ), searchAlgorithm( NULL ), useDistObj( -1 )
{
    complexity.clear();
    geomType.clear();
    fileNames.clear();
    objects.clear();
    algorithms.clear();
}

/*! ----------------------------------------------------------------------------
 *  \brief  creates settinggs for benchmarkfrom command line
 *  \param argv command line
 *  \param argc #command line parameters
 ------------------------------------------------------------------------------- */
Settings::Settings( int argc, char *argv[] ) : Obj( 0 ), algo( 0 ), curDist( 0 ), curFile( 0 ), saveName( 0 ), qridSize( 0 ), method( METH_SPHERE ), dist( -1 ), rotVel( 60 ), degVel( 30 ), validate( false ), noTiming( false ), withWindow( false ), whiteBackground( false ), choice( 0 ), maxminTiming( true ), searchAlgorithm( NULL ), useDistObj( -1 )
{
    complexity.clear();
    geomType.clear();
    fileNames.clear();
    objects.clear();
    algorithms.clear();
    parsecommandline( argc, argv );
}

/* -----------------------------------------------------------------------------
 *  \brief resets current tested distance
 ------------------------------------------------------------------------------- */
void Settings::resetDists()
{
    curDist = 0;
}

/* -----------------------------------------------------------------------------
 *  \brief returns next distance to test
 ------------------------------------------------------------------------------- */
bool Settings::nextDist()
{
    if( curDist >= dists.size() )
    {
        curDist = 0;
        return true;
    }

    dist = dists[curDist++];
    return false;
}

/* -----------------------------------------------------------------------------
 *  \brief returns name of current tested configuration file
 ------------------------------------------------------------------------------- */
const char *Settings::getLoadname()
{
    if( curFile >= loadNames.size() )
    {
        curFile = 0;
        return 0;
    }

    return loadNames[curFile++];
}

/* -----------------------------------------------------------------------------
 *  \brief reset current algorithm
 ------------------------------------------------------------------------------- */
void Settings::resetAlgorithms()
{
    algo = 0;
}

/* -----------------------------------------------------------------------------
 *  \brief returns current tested object
 ------------------------------------------------------------------------------- */
osg::NodePtr Settings::getObject()
{
    if( Obj >= objects.size() )
        objects.push_back( objects[Obj - 1]->clone() );
    return objects[Obj++];
}


/* -----------------------------------------------------------------------------
 *  \brief returns first fixed object
 ------------------------------------------------------------------------------- */
osg::NodePtr Settings::getFixedNode()
{
    if( objects.size() > 0 )
        return objects[0];
    return osg::NullFC;
}

/* -----------------------------------------------------------------------------
 *  \brief returns first moving object
 ------------------------------------------------------------------------------- */
osg::NodePtr Settings::getMovingNode()
{
    if( objects.size() > 0 )
        return objects[1];
    return osg::NullFC;
}



/* -----------------------------------------------------------------------------
 *  \brief return current algorithm
 ------------------------------------------------------------------------------- */
const Algorithm *Settings::getAlgorithm()
{
    if( algo >= algorithms.size() )
        return 0;
    return algorithms[algo++];
}

/*! ----------------------------------------------------------------------------
 *  \brief  Creates objects from setting-parameters Loads objects from files or
 *          creates standard OpenSG objects due to parameters set in the
 *          settings-object
 ------------------------------------------------------------------------------- */
void Settings::buildObjects()
{
    int Fn = 0, Cl = 0;

    if( !geomType.size() )
        geomType.push_back( OBJ_PLANES );
    if( !complexity.size() )
        complexity.push_back( 3 );
    for( int i = 0; i < geomType.size(); ++i )
    {
        if( geomType[i] == OBJ_FILE )
        {
            if( !fileNames.size() )
            {
                fputs( "bench: '-g file' option given, but moving node or fixed node is still NULL!",
                   stderr );
                exit( -1 );
            }

            char *filename = (char *) fileNames[Fn];//rwrw
            objects.push_back( loadGeom( fileNames[Fn] ) );
            if( Fn < fileNames.size() - 1 )
                ++Fn;
        }
        else
        {
            objects.push_back( createGeom( geomType[i], complexity[Cl] ) );
            if( Cl < complexity.size() - 1 )
                ++Cl;
        }
    }
}

/*! ----------------------------------------------------------------------------
 *  \brief  Called when command line-parameters or options are wrong
 *  \param parm indicates wrong command-line parameter 
 *  \param cmd indicates wrong command-line option
 ------------------------------------------------------------------------------- */
void Settings::commandlineerror( char *cmd, char *parm )
{
    std::list<const char *> suffixes;
    if( cmd )
        fprintf( stderr, "Offending option: %s\n", cmd );
    if( parm )
        fprintf( stderr, "with first parameter: %s\n", parm );

    fprintf( stderr, "\n\nUsage: bench options ...\n""Options:\n"
		"-x compl    complexity (#pgons ~ 2*compl^2)\n"
        "-g obj      geometry type (default = pl)\n"
        "              obj = pl, sh, to;\n"
        "              if obj='file', then the object loaded with option -f is used\n"
        "-r deg      rotate moving object about deg degrees (default=60)\n"
        "-a deg      move object by deg degrees on orbit sphere (default=30)\n"
        "-d d,..     compute positions such that objects have a distance of d,\n"
		"              negative values result in collisions (default=-1)\n"
        "-D obj      use offset surfaces to compute distance between objs\n"
		"              0 = don't, 1 = low detail, 2 = high detail\n"
        "-F algo     if '-D', algorithm to use to coll.det. with offset objs\n"
        "-A algo,..  algorithm to use for coll. det. (default = %s)\n",
        algorithm ? algorithm->shortName : "None" );
        for( Algorithm * a = algorithm; a; a = a->next )
        {
            fprintf( stderr, "              %s = %s\n", a->shortName, a->name );
        }

        fprintf( stderr,
		"-f file     load file and use the node with name 'benchobj'\n"
		"-p meth     method for finding positions (default=sphere)\n"
        "              meth = sphere, csphere, grid\n"
        "-G x[,y[,z]]qridSize (default=20)\n"
        "-T          do not print timing info (for regression tests)\n"
        "-V          verify, print number of collisions\n"
        "-N          disable maximum/minimum timing\n"
        "-w          with window; timing will be bogus\n"
        "            (program won't stop automatically)\n"
        "-W          white background (default = black)\n"
        "-s name     save collisionpoints to file name\n"
        "-l name1,.. load collisionpoints from file name1 ...\n"
        "-h          this help menu\n"
        "Keys:\n"
        "l           switch lighting mode\n"
        "p           switch drawing mode (filled/wireframe/point)\n"
        "r           reset the list of previous positions\n"
        "s	         show closeste points of moving objct\n"
        "a           stop/restart configuration computation when finished\n"
        "m           grab screenshots in ./movie-directory\n"
        "t           slow motion in movie\n"
        "u	         turns on speed up in movie recording mode\n"
        "i           increase speed up in movie recording mode\n"
        "[space]     pause\n""[enter]     step\n""q           quit\n"
        "\n" );

    puts( "Supported file formats:" );
    osg::SceneFileHandler::the().getSuffixList( suffixes );
    for( std::list < const char * >::iterator it = suffixes.begin();
         it != suffixes.end(); ++it )
    {
        fputs( *it, stdout );
        fputs( "  ", stdout );
    }

    putchar( '\n' );

    if( cmd )
        exit( -1 ); /* problem occured */
    else
        exit( 0 );  /* was option -h */
}

/*! ----------------------------------------------------------------------------
 *  \brief  Display current tested configuration on command-line
 ------------------------------------------------------------------------------- */
void Settings::printHeader()
{
    printf( "dist" );
    resetAlgorithms();
    for( const Algorithm * A = getAlgorithm(); A; A = getAlgorithm() )
    {
        printf( "\t%s", A->name );
    }

    printf( "\n" );
}


/*! ----------------------------------------------------------------------------
 *  \brief  Reads options and parameters from the command line
 *  \param argc #command line options 
 *  \param argv command line
 ------------------------------------------------------------------------------- */
void Settings::parsecommandline( int argc, char *argv[] )
{

    /*
     * valid option characters;
     * last char MUST be 0 !
     */
    char    optionchar[] =    { 'h', 'g', 'x', 'p', 'r', 'a', 'w', 'd', 'A', 'T', 'f', 'W', 's', 'l', 'V', 'G', 'N', 'D', 'F', 0 };
    int     musthaveparam[] = {  0,   1,   1,   1,   1,   1,   0,   1,   1,   0,   1,   0,   1,   1,   0,   1,   0,   1,   1,  0 };
    char    *tmp;
    int     nopts;
    int     mhp[256] = { 0 };
    int     isopt[256] = { 0 };
    int     optchar;

    nopts = strlen( optionchar );
    if( nopts > 50 )
    {
        fprintf( stderr, "\n\nparsecommandline: the option-chars string "
                 "seems to be\nVERY long (%d bytes) !\n\n", nopts );
        exit( -1 );
    }

    for( int i = 0; i < nopts; i++ )
    {
        if( isopt[int( optionchar[i] )] )
        {
            fprintf( stderr, "\n\nparsecommandline: Bug: option character '%c'"
                     " is specified twice in the\n"
                 "option character array !\n\n", optionchar[i] );
            exit( -1 );
        }

        isopt[int( optionchar[i] )] = 1;
        mhp[int( optionchar[i] )] = musthaveparam[i];
    }

    ++argv;
    --argc;
    while( argc > 0 )
    {
        if( argv[0][0] == '-' )
        {
            optchar = argv[0][1];

            if( !isopt[optchar] )
            {
                fprintf( stderr, "\nIs not a valid command line option\n" );
                commandlineerror( argv[0], NULL );
            }

            for( int i = 0; i < mhp[optchar]; i++ )
            {
                if( !argv[1 + i] || argv[1 + i][0] == '-' )
                {
                    fprintf( stderr, "\nCommand line option -%c must "
                             "have %d parameter(s)\n", argv[0][1],
                             mhp[optchar] );
                    commandlineerror( argv[0], NULL );
                    argv += 1 + i;
                    argc -= 1 + i;
                    continue;
                }
            }

            switch( optchar )
            {
                case 'h':
                    commandlineerror( NULL, NULL );
                    break;

                case 'w':
                    withWindow = true;
                    break;

                case 'W':
                    whiteBackground = true;
                    break;

                case 'V':
                    validate = true;
                    break;

                case 'T':
                    noTiming = true;
                    break;
                    
                case 'N':
                    maxminTiming = false;
                    break;

                case 'x':
                    while( *(tmp = getToken( &argv[1], "," )) )
                    {
                        int comp = atoi( tmp );
                        if( comp < 3 || comp > 1000 )
                        {
                            fprintf( stderr, "complexity (%d) out of range!\n", comp );
                            exit( 1 );
                        }

                        complexity.push_back( comp );
                    }
                    break;

                case 'd':
                    while( *(tmp = getToken( &argv[1], "," )) )
                    {
                        dists.push_back( atof( tmp ) );
                    }

                    dist = dists[0];
                    break;

                case 'r':
                    rotVel = atof( argv[1] );
                    break;

                case 'a':
                    degVel = atof( argv[1] );
                    break;

                case 'g':
                    while( *(tmp = getToken( &argv[1], "," )) )
                    {
                        if( !strcmp( tmp, "pl" ) )
                            geomType.push_back( OBJ_PLANES );
                        else if( !strcmp( tmp, "sh" ) )
                            geomType.push_back( OBJ_SPHERES );
                        else if( !strcmp( tmp, "to" ) )
                            geomType.push_back( OBJ_TORUS );
                        else if( !strcmp( tmp, "file" ) )
                            geomType.push_back( OBJ_FILE );
                        else
                        {
                            fputs( "unrecognized obj type\n", stderr );
                            commandlineerror( argv[0], tmp );
                        }
                    }
                    break;

                case 'p':
                    if( !strcmp( argv[1], "sphere" ) )
                        method = METH_SPHERE;
                    else if( !strcmp( argv[1], "csphere" ) )
                        method = METH_CHECKSPHERE;
                    else if( !strcmp( argv[1], "grid" ) )
                        method = METH_GRID;
                    else
                    {
                        fputs( "unrecognized meth type\n", stderr );
                        commandlineerror( argv[0], argv[1] );
                    }
                    break;

                case 'A':
                    while( *(tmp = getToken( &argv[1], "," )) )
                    {
                        choice = 0;
                        for( Algorithm * a = algorithm; a; a = a->next )
                        {
                            if( !strcmp( tmp, a->shortName ) )
                            {
                                choice = a;
                                break;
                            }
                        }

                        if( !choice )
                        {
                            fputs( "unrecognized algorithm\n", stderr );
                            commandlineerror( argv[0], tmp );
                        }

                        algorithms.push_back( choice );
                    }
                    break;

                case 'F':
                    choice = 0;
                    for( Algorithm * a = algorithm; a; a = a->next )
                    {
                        if( !strcmp( argv[1], a->shortName ) )
                        {
                            searchAlgorithm = a;
                            break;
                        }
                    }

                    if( !searchAlgorithm )
                    {
                        fputs( "unrecognized algorithm\n", stderr );
                        commandlineerror( argv[0], argv[1] );
                    }
                    break;

                case 'f':
                    while( *(tmp = getToken( &argv[1], "," )) )
                    {
                        fileNames.push_back( tmp );
                    }
                    break;

                case 's':
                    saveName = argv[1];
                    break;

                case 'G':
                    qridSize = argv[1];
                    break;

                case 'l':
                    while( *(tmp = getToken( &argv[1], "," )) )
                    {
                        loadNames.push_back( tmp );
                    }
                    break;

                case 'D':
                    useDistObj = atoi(argv[1]);
                    if (useDistObj < 0 || useDistObj > 2)
                    {
                        fputs( "unrecognized distance object\n", stderr );
                        commandlineerror( argv[0], argv[1] );
                    }
                    break;

                default:
                    fprintf( stderr, "\nBug in parsecommandline !\n" );
                    commandlineerror( argv[0], NULL );
            }

            argv += 1 + mhp[optchar];
            argc -= 1 + mhp[optchar];
        }
        else
        {

            /* command line arg doesn't start with '-' */
            fprintf( stderr, "\nThis is not a valid command line option\n" );
            commandlineerror( argv[0], NULL );

            /* or, load file instead .. */
        }
    }

    /* check sanity of options */
    if( dists.size() == 0 && loadNames.size() == 0 )
        dists.push_back( -1 );

    for( Algorithm * a = algorithm; a; a = a->next )
    {
        if( !choice && a->api.doCollisions )
            choice = a;
        if( !searchAlgorithm && a->api.getDistance )
            searchAlgorithm = a;
    }
    if (!searchAlgorithm) searchAlgorithm = choice;
    if (!searchAlgorithm)
    {
        fputs( "no algorithms available.\n", stderr );
        commandlineerror( argv[0], NULL );
    }

    if ( useDistObj < 0 )
    {
        useDistObj = !searchAlgorithm->api.getDistance;
    }

    if( !algorithms.size() )
        algorithms.push_back( choice );

    buildObjects();
}
